/*
 * TOFU Circuit Simulator
 * Copyright (C) 2007 Michael Birken
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
 * 
 * http://www.meatfighter.com/tofu
 */

package tofu.harryporter;

import java.io.*;
import java.util.*;

/* 
 * MOV D S   (A,B,C,D,M1,M2,X,Y)  1 byte 
 * CLR R     (A,B,C,D,M1,M2,X,Y)  1 byte
 * LD R      (A,B,C,D)            1 byte
 * STO R     (A,B,C,D)            1 byte
 * ADD R     (A,D)                1 byte 
 * INC R     (A,D,XY)             1 byte  
 * AND R     (A,D)                1 byte 
 * OR R      (A,D)                1 byte 
 * XOR R     (A,D)                1 byte 
 * NOT R     (A,D)                1 byte 
 * SHL R     (A,D)                1 byte  
 * JMP label                      3 bytes
 * CAL label                      3 bytes
 * RET                            1 byte
 * BE/BZ label                    3 bytes
 * BNE/BNZ label                  3 bytes
 * BNC label                      3 bytes
 * BNG label                      3 bytes
 * SET M/J 16bits                 3 bytes
 * SET R 5bits (A,B)              1 byte
 * HLT                            1 byte
 *
 * MOV PC XY
 * MOV PC M
 * MOV PC J
 * MOV XY M
 * MOV XY J
 *
 * label:
 * DATA hex pairs                 x bytes
 * WORD word                      2 bytes
 * BYTE byte                      1 byte
 * STRING "..."                   x bytes
 */

public class Assembler {
  
  private enum Operation {
    MOVE, CLEAR, LOAD, STORE, ADD, INC, AND, OR, XOR, NOT, SHIFT_LEFT, GOTO, 
    CALL, RETURN, BRANCH_EQUAL, BRANCH_NOT_EQUAL, BRANCH_NOT_CARRY, 
    BRANCH_NEGATIVE, SET16, SET5, HALT, DATA, WORD, BYTE
  } 
  
  private enum Register { A, B, C, D, M1, M2, X, Y, XY, PC, M, J, ZERO }
  
  private class Instruction {
    private int lineNumber;
    private Operation operation;
    private Register register1;
    private Register register2;
    private int immediate;
    private String label;
    private int[] bytes;
    
    public Instruction(Operation operation) {
      this.operation = operation;
    }
    
    public Instruction(Operation operation, Register register) {
      this.operation = operation;
      this.register1 = register;
    }
    
    public Instruction(
        Operation operation, Register register1, Register register2) {
      this.operation = operation;
      this.register1 = register1;
      this.register2 = register2;
    }
    
    public Instruction(Operation operation, String label, int lineNumber) {
      this.lineNumber = lineNumber;
      this.operation = operation;
      this.label = label;
    }    
    
    public Instruction(
        Operation operation, Register register, String label, int lineNumber) {
      this.lineNumber = lineNumber;
      this.operation = operation;
      this.register1 = register;
      this.label = label;
    }
    
    public Instruction(Operation operation, int immediate) {
      this.operation = operation;
      this.immediate = immediate;
    }
    
    public Instruction(Operation operation, Register register, int immediate) {
      this.operation = operation;
      this.register1 = register;
      this.immediate = immediate;
    }    
    
    public Instruction(Operation operation, int[] bytes) {
      this.operation = operation;
      this.bytes = bytes;
    }
  }
  
  private Map<String, Integer> labelMap = new HashMap<String, Integer>();
  private List<Instruction> instructions = new ArrayList<Instruction>();
  private int programCounter;
  
  private void compile(String fileName) throws IOException {
    
    BufferedOutputStream bos = null;
    try {
      bos = new BufferedOutputStream(new FileOutputStream(fileName)) {
        private int i = 0;
        public void write(int b) throws IOException {
          super.write(b);          
          int x = (b < 0) ? 256 + (int)b : (int)b;
          i++;
        }          
      };
      for(int i = 0; i < instructions.size(); i++) {
        if (!writeInstruction(bos, instructions.get(i))) {
          return;
        }
      }
    } finally {
      if (bos != null) {
        bos.close();
      }
    }
  }
  
  private boolean writeInstruction(
      OutputStream out, Instruction instruction) throws IOException {
    Integer label = null;
    switch(instruction.operation) {
      case MOVE:
        out.write((translateRegister(instruction.register1) << 3)
            | translateRegister(instruction.register2));
        break;
      case CLEAR:
        out.write((translateRegister(instruction.register1) << 3)
            | translateRegister(instruction.register1));       
        break;
      case LOAD:
        out.write(0x90 | translateRegister(instruction.register1));
        break;
      case STORE:
        out.write(0x98 | translateRegister(instruction.register1));
        break;  
      case ADD:
        out.write(instruction.register1 == Register.A ? 0x80 : 0x88);
        break;
      case INC:
        switch(instruction.register1) {
          case A:
            out.write(0x81);
            break;
          case D:
            out.write(0x89);
            break;
          case XY:
            out.write(0xB0);
            break;
        }
        break;
      case AND:
        out.write(instruction.register1 == Register.A ? 0x82 : 0x8A);
        break;
      case OR:
        out.write(instruction.register1 == Register.A ? 0x83 : 0x8B);
        break;
      case XOR:
        out.write(instruction.register1 == Register.A ? 0x84 : 0x8C);
        break;
      case NOT:
        out.write(instruction.register1 == Register.A ? 0x85 : 0x8D);
        break;
      case SHIFT_LEFT:
        out.write(instruction.register1 == Register.A ? 0x86 : 0x8E);
        break;
      case GOTO:
        out.write(0xE6);
        if ((label = labelMap.get(instruction.label)) == null) {
          printError(instruction.lineNumber, 
              "Cannot find label: " + instruction.label);
          return false;
        }
        out.write(0xFF & (label >> 8));
        out.write(0xFF & label);
        break;
      case CALL:
        out.write(0xE7);
        if ((label = labelMap.get(instruction.label)) == null) {
          printError(instruction.lineNumber, 
              "Cannot find label: " + instruction.label);
          return false;
        }
        out.write(0xFF & (label >> 8));
        out.write(0xFF & label);
        break;  
      case RETURN:
        out.write(instruction.register1 == Register.XY ?
              0xA0 | translateRegister(instruction.register2)
            : 0xA8 | translateRegister(instruction.register2));
        break;
      case BRANCH_EQUAL:
        out.write(0xE4);
        if ((label = labelMap.get(instruction.label)) == null) {
          printError(instruction.lineNumber, 
              "Cannot find label: " + instruction.label);
          return false;
        }
        out.write(0xFF & (label >> 8));
        out.write(0xFF & label);
        break;
      case BRANCH_NOT_EQUAL:
        out.write(0xE2);
        if ((label = labelMap.get(instruction.label)) == null) {
          printError(instruction.lineNumber, 
              "Cannot find label: " + instruction.label);
          return false;
        }
        out.write(0xFF & (label >> 8));
        out.write(0xFF & label);
        break;
      case BRANCH_NOT_CARRY:
        out.write(0xE8);
        if ((label = labelMap.get(instruction.label)) == null) {
          printError(instruction.lineNumber, 
              "Cannot find label: " + instruction.label);
          return false;
        }
        out.write(0xFF & (label >> 8));
        out.write(0xFF & label);
        break;
      case BRANCH_NEGATIVE:
        out.write(0xF0);
        if ((label = labelMap.get(instruction.label)) == null) {
          printError(instruction.lineNumber, 
              "Cannot find label: " + instruction.label);
          return false;
        }
        out.write(0xFF & (label >> 8));
        out.write(0xFF & label);
        break;
      case SET16:
        if (instruction.register1 == Register.M) {
          out.write(0xC0);
        } else {
          out.write(0xE0);
        }
        if (instruction.label == null) {
          out.write(0xFF & (instruction.immediate >> 8));
          out.write(0xFF & instruction.immediate);
        } else {
          if ((label = labelMap.get(instruction.label)) == null) {
            printError(instruction.lineNumber, 
                "Cannot find label: " + instruction.label);
            return false;
          }
          out.write(0xFF & (label >> 8));
          out.write(0xFF & label);
        }
        break;
      case SET5:
        out.write((instruction.immediate & 0x1F)
            | (instruction.register1 == Register.A ? 0x40 : 0x60));
        break;
      case HALT:
        out.write(0xAE);
        break;
      case DATA:
        for(int i = 0; i < instruction.bytes.length; i++) {
          out.write(instruction.bytes[i]);
        }
        break;
      case WORD:
        out.write(0xFF & (instruction.immediate >> 8));
        out.write(0xFF & instruction.immediate);
        break;
      case BYTE:
        out.write(0xFF & instruction.immediate);
        break;        
    }
    return true;
  }
  
  private int translateRegister(Register register) {
    switch(register) {
      case A:
        return 0;
      case B:
        return 1;
      case C:
        return 2;
      case D:
        return 3;
      case M1:
        return 4;
      case M2:
        return 5;
      case X:
        return 6;
      case Y:
        return 7;
      case M:
        return 0;
      case XY:
        return 2;
      case J:
        return 4;
      case ZERO:
        return 6;
    }
    return 0;
  }
  
  private boolean processFileLine(int lineNumber, String line) {  
    int pivot = line.indexOf("#");
    if (pivot == 0) {
      return true;
    } else if (pivot > 0) {
      String tmp = line.trim().toUpperCase();
      if (tmp.startsWith("STRING")) {
        int index = tmp.lastIndexOf("\"");
        pivot = line.indexOf("#", index);
        if (pivot > 0) {
          line = line.substring(0, pivot);
        }
      } else {
        line = line.substring(0, pivot);
      }
    }
    line = line.trim();
    String originalCaseLine = line;
    line = line.toUpperCase();
    if (line.length() == 0) {
      return true;
    }
    String[] tokens = line.split("\\s+");
    String instruction = tokens[0];
    if ("MOV".equals(instruction)) {
      if (tokens.length != 3) {
        printError(lineNumber, 
            "Syntax: MOV A|B|C|D|M1|M2|X|Y A|B|C|D|M1|M2|X|Y\n" +
            "        MOV PC XY\n" +
            "        MOV PC M\n" +
            "        MOV PC J\n" +
            "        MOV XY M\n" +
            "        MOV XY J");
        return false;
      }
      Register register1 = "A".equals(tokens[1]) ? Register.A 
                         : "B".equals(tokens[1]) ? Register.B
                         : "C".equals(tokens[1]) ? Register.C
                         : "D".equals(tokens[1]) ? Register.D
                         : "M1".equals(tokens[1]) ? Register.M1
                         : "M2".equals(tokens[1]) ? Register.M2
                         : "X".equals(tokens[1]) ? Register.X
                         : "Y".equals(tokens[1]) ? Register.Y 
                         : "PC".equals(tokens[1]) ? Register.PC
                         : "XY".equals(tokens[1]) ? Register.XY
                         : null;
      if (register1 == null) {
        printError(lineNumber, 
            "Syntax: MOV A|B|C|D|M1|M2|X|Y A|B|C|D|M1|M2|X|Y\n" +
            "        MOV PC XY\n" +
            "        MOV PC M\n" +
            "        MOV PC J\n" +
            "        MOV XY M\n" +
            "        MOV XY J");
        return false;
      }
      Register register2 = "A".equals(tokens[2]) ? Register.A 
                         : "B".equals(tokens[2]) ? Register.B
                         : "C".equals(tokens[2]) ? Register.C
                         : "D".equals(tokens[2]) ? Register.D
                         : "M1".equals(tokens[2]) ? Register.M1
                         : "M2".equals(tokens[2]) ? Register.M2
                         : "X".equals(tokens[2]) ? Register.X
                         : "Y".equals(tokens[2]) ? Register.Y 
                         : "XY".equals(tokens[2]) ? Register.XY
                         : "M".equals(tokens[2]) ? Register.M
                         : "J".equals(tokens[2]) ? Register.J
                         : null;
      if (register2 == null) {
        printError(lineNumber, 
            "Syntax: MOV A|B|C|D|M1|M2|X|Y A|B|C|D|M1|M2|X|Y\n" +
            "        MOV PC XY\n" +
            "        MOV PC M\n" +
            "        MOV PC J\n" +
            "        MOV XY M\n" +
            "        MOV XY J");
        return false;
      }   
      if ((register1 == Register.PC 
          && register2 != Register.XY
          && register2 != Register.M
          && register2 != Register.J) ||
          (register1 == Register.XY
          && register2 != Register.M
          && register2 != Register.J)) {
        printError(lineNumber, 
            "Syntax: MOV A|B|C|D|M1|M2|X|Y A|B|C|D|M1|M2|X|Y\n" +
            "        MOV PC XY\n" +
            "        MOV PC M\n" +
            "        MOV PC J\n" +
            "        MOV XY M\n" +
            "        MOV XY J");
        return false;        
      }
      if (register1 == Register.PC || register1 == Register.XY) {
        instructions.add(new Instruction(
            Operation.RETURN, register1, register2));  
      } else { 
        instructions.add(new Instruction(
            Operation.MOVE, register1, register2));
      }
      programCounter++;
    } else if ("CLR".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: CLR A|B|C|D|M1|M2|X|Y");
        return false;
      }
      Register register = "A".equals(tokens[1]) ? Register.A 
                        : "B".equals(tokens[1]) ? Register.B
                        : "C".equals(tokens[1]) ? Register.C
                        : "D".equals(tokens[1]) ? Register.D
                        : "M1".equals(tokens[1]) ? Register.M1
                        : "M2".equals(tokens[1]) ? Register.M2
                        : "X".equals(tokens[1]) ? Register.X
                        : "Y".equals(tokens[1]) ? Register.Y 
                        : null;
      if (register == null) {
        printError(lineNumber, "Syntax: CLR A|B|C|D|M1|M2|X|Y");
        return false;
      }
      instructions.add(new Instruction(
          Operation.CLEAR, register));
      programCounter++;
    } else if ("LD".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: LD A|B|C|D");
        return false;
      }
      Register register = "A".equals(tokens[1]) ? Register.A 
                        : "B".equals(tokens[1]) ? Register.B
                        : "C".equals(tokens[1]) ? Register.C
                        : "D".equals(tokens[1]) ? Register.D
                        : null;
      if (register == null) {
        printError(lineNumber, "Syntax: LD A|B|C|D");
        return false;
      }
      instructions.add(new Instruction(
          Operation.LOAD, register));
      programCounter++;      
    } else if ("STO".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: STO A|B|C|D");
        return false;
      }
      Register register = "A".equals(tokens[1]) ? Register.A 
                        : "B".equals(tokens[1]) ? Register.B
                        : "C".equals(tokens[1]) ? Register.C
                        : "D".equals(tokens[1]) ? Register.D
                        : null;
      if (register == null) {
        printError(lineNumber, "Syntax: STO A|B|C|D");
        return false;
      }
      instructions.add(new Instruction(Operation.STORE, register));
      programCounter++;  
    } else if ("ADD".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: ADD A|D");
        return false;
      }
      Register register = "A".equals(tokens[1]) ? Register.A
                        : "D".equals(tokens[1]) ? Register.D
                        : null;
      if (register == null) {
        printError(lineNumber, "Syntax: ADD A|D");
        return false;
      }      
      instructions.add(new Instruction(Operation.ADD, register));
      programCounter++;  
    } else if ("INC".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: INC A|D|XY");
        return false;
      }
      Register register = "A".equals(tokens[1]) ? Register.A
                        : "D".equals(tokens[1]) ? Register.D
                        : "XY".equals(tokens[1]) ? Register.XY
                        : null;
      if (register == null) {
        printError(lineNumber, "Syntax: INC A|D|XY");
        return false;
      }
      instructions.add(new Instruction(Operation.INC, register));
      programCounter++;  
    } else if ("AND".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: AND A|D");
        return false;
      }
      Register register = "A".equals(tokens[1]) ? Register.A
                        : "D".equals(tokens[1]) ? Register.D
                        : null;
      if (register == null) {
        printError(lineNumber, "Syntax: AND A|D");
        return false;
      }      
      instructions.add(new Instruction(Operation.AND, register));
      programCounter++;  
    } else if ("OR".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: OR A|D");
        return false;
      }
      Register register = "A".equals(tokens[1]) ? Register.A
                        : "D".equals(tokens[1]) ? Register.D
                        : null;
      if (register == null) {
        printError(lineNumber, "Syntax: OR A|D");
        return false;
      }      
      instructions.add(new Instruction(Operation.OR, register));
      programCounter++;  
    } else if ("XOR".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: XOR A|D");
        return false;
      }
      Register register = "A".equals(tokens[1]) ? Register.A
                        : "D".equals(tokens[1]) ? Register.D
                        : null;
      if (register == null) {
        printError(lineNumber, "Syntax: XOR A|D");
        return false;
      }      
      instructions.add(new Instruction(Operation.XOR, register));
      programCounter++;  
    } else if ("NOT".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: NOT A|D");
        return false;
      }
      Register register = "A".equals(tokens[1]) ? Register.A
                        : "D".equals(tokens[1]) ? Register.D
                        : null;
      if (register == null) {
        printError(lineNumber, "Syntax: NOT A|D");
        return false;
      }      
      instructions.add(new Instruction(Operation.NOT, register));
      programCounter++;  
    } else if ("SHL".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: SHL A|D");
        return false;
      }
      Register register = "A".equals(tokens[1]) ? Register.A
                        : "D".equals(tokens[1]) ? Register.D
                        : null;
      if (register == null) {
        printError(lineNumber, "Syntax: SHL A|D");
        return false;
      }      
      instructions.add(new Instruction(Operation.SHIFT_LEFT, register));
      programCounter++;  
    } else if ("JMP".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: JMP label");
        return false;
      }
      instructions.add(new Instruction(Operation.GOTO, tokens[1], lineNumber));      
      programCounter += 3;
    } else if ("CAL".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: CAL label");
        return false;
      }
      instructions.add(new Instruction(Operation.CALL, tokens[1], lineNumber));      
      programCounter += 3;
    } else if ("RET".equals(instruction)) {
      if (tokens.length != 1) {
        printError(lineNumber, "Syntax: RET");
        return false;
      }
      Register register1 = Register.PC;
      Register register2 = Register.XY;     
      instructions.add(new Instruction(Operation.RETURN, register1, register2));
      programCounter++;  
    } else if ("BE".equals(instruction) || "BZ".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: BE/BZ label");
        return false;
      }
      instructions.add(new Instruction(
          Operation.BRANCH_EQUAL, tokens[1], lineNumber));      
      programCounter += 3;
    } else if ("BNE".equals(instruction) || "BNZ".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: BNE/BNZ label");
        return false;
      }
      instructions.add(new Instruction(
          Operation.BRANCH_NOT_EQUAL, tokens[1], lineNumber));      
      programCounter += 3;
    } else if ("BNC".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: BNC label");
        return false;
      }
      instructions.add(new Instruction(
          Operation.BRANCH_NOT_CARRY, tokens[1], lineNumber));      
      programCounter += 3;
    } else if ("BNG".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: BNG label");
        return false;
      }
      instructions.add(new Instruction(
          Operation.BRANCH_NEGATIVE, tokens[1], lineNumber));      
      programCounter += 3;
    } else if ("SET".equals(instruction)) {
      if (tokens.length != 3) {
        printError(lineNumber, 
            "Syntax: (SET M|J 16bits)|(SET M|J label)|(SET A|B 5bits)");
        return false;
      }
      if ("M".equals(tokens[1]) || "J".equals(tokens[1])) {
        int immediate = 0;
        boolean isLabel = false;
        if (tokens[2].endsWith("H")) {          
          try {
            immediate = Integer.parseInt(
                tokens[2].substring(0, tokens[2].length() - 1), 16);
          } catch(NumberFormatException e) {
            isLabel = true;
          }
        } else {
          try {
            immediate = Integer.parseInt(tokens[2]);
          } catch(NumberFormatException e) {
            isLabel = true;
          }
        }
        Register register = "M".equals(tokens[1]) ? Register.M : Register.J;
        if (isLabel) {
          instructions.add(new Instruction(
              Operation.SET16, register, tokens[2], lineNumber)); 
        } else {
          instructions.add(new Instruction(
              Operation.SET16, register, immediate));  
        }
        programCounter += 3;
      } else if ("A".equals(tokens[1]) || "B".equals(tokens[1])) {
        Register register = "A".equals(tokens[1]) ? Register.A : Register.B;
        int immediate = 0;
        if (tokens[2].endsWith("H")) {          
          try {
            immediate = Integer.parseInt(
                tokens[2].substring(0, tokens[2].length() - 1), 16);
          } catch(NumberFormatException e) {
            printError(lineNumber, "Invalid hex immediate value.");
            return false;
          }
        } else {
          try {
            immediate = Integer.parseInt(tokens[2]);
          } catch(NumberFormatException e) {
            printError(lineNumber, "Invalid decimal immediate value.");
            return false;
          }
        }
        instructions.add(new Instruction(Operation.SET5, register, immediate));      
        programCounter++;        
      } else {
        printError(lineNumber, 
            "Syntax: (SET M|J 16bits)|(SET M|J label)|(SET A|B 5bits)");
        return false;
      }
    } else if ("HLT".equals(instruction)) {
      if (tokens.length != 1) {
        printError(lineNumber, "Syntax: HLT");
        return false;
      }
      instructions.add(new Instruction(Operation.HALT));
      programCounter++;  
    } else if (instruction.endsWith(":")) {
      if (tokens.length != 1 || instruction.length() < 2) {
        printError(lineNumber, "Invalid label.");
        return false;
      }
      labelMap.put(instruction.substring(0, instruction.length() - 1), 
          programCounter);
    } else if ("DATA".equals(instruction)) {
      if (tokens.length == 1) {
        printError(lineNumber, "Syntax: DATA HH HH HH ... HH");
        return false;
      }
      int[] bytes = new int[tokens.length - 1];
      for(int i = 1; i < tokens.length; i++) {
        try {
          bytes[i - 1] = Integer.parseInt(tokens[i], 16);
        } catch(NumberFormatException e) {
          printError(lineNumber, "Invalid hex pair: " + tokens[i]);
          return false;
        }
        if ((bytes[i - 1] & 0xFFFFFF00) != 0) {
          printError(lineNumber, "Not hex pair: " + tokens[i]);
          return false;
        }        
      }
      instructions.add(new Instruction(Operation.DATA, bytes));
      programCounter += bytes.length;
    } else if ("WORD".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: WORD word");
        return false;
      }
      int immediate = 0;
      if (tokens[1].endsWith("H")) {          
        try {
          immediate = Integer.parseInt(
              tokens[1].substring(0, tokens[1].length() - 1), 16);
        } catch(NumberFormatException e) {
          printError(lineNumber, "Invalid hex value.");
          return false;
        }
      } else {
        try {
          immediate = Integer.parseInt(tokens[1]);
        } catch(NumberFormatException e) {
          printError(lineNumber, "Invalid decimal value.");
          return false;
        }
      }     
      instructions.add(new Instruction(Operation.WORD, immediate));
      programCounter += 2;
    } else if ("BYTE".equals(instruction)) {
      if (tokens.length != 2) {
        printError(lineNumber, "Syntax: BYTE byte");
        return false;
      }
      int immediate = 0;
      if (tokens[1].endsWith("H")) {          
        try {
          immediate = Integer.parseInt(
              tokens[1].substring(0, tokens[1].length() - 1), 16);
        } catch(NumberFormatException e) {
          printError(lineNumber, "Invalid hex value.");
          return false;
        }
      } else {
        try {
          immediate = Integer.parseInt(tokens[1]);
        } catch(NumberFormatException e) {
          printError(lineNumber, "Invalid decimal value.");
          return false;
        }
      }     
      instructions.add(new Instruction(Operation.BYTE, immediate));
      programCounter++; 
    } else if ("STRING".equals(instruction)) {
      int start = originalCaseLine.indexOf('"');
      int end = originalCaseLine.lastIndexOf('"');
      if (start < 0 || end < 0 || start == end) {
        printError(lineNumber, "Syntax: STRING \"...\"");
        return false;
      }
      String str = originalCaseLine.substring(start + 1, end);
      int[] bytes = new int[str.length()];
      for(int i = 0; i < str.length(); i++) {
        bytes[i] = str.charAt(i);
      }
      instructions.add(new Instruction(Operation.DATA, bytes));
      programCounter += bytes.length;
    } else {
      printError(lineNumber, "Unknown command.");
      return false;
    }
    return true;
  }
  
  private void printError(int line, String message) {
    System.out.printf("%d: %s%n", line, message);
  }
  
  private boolean readFile(String fileName) throws IOException {
    BufferedReader br = null;
    int lineNumber = 0;
    try {
      br = new BufferedReader(new FileReader(fileName));
      String input = null;
      while((input = br.readLine()) != null) {
        if (!processFileLine(++lineNumber, input)) {
          return false;
        }
      }
    } finally {
      if (br != null) {
        try {
          br.close();
        } catch(IOException e) {          
        }
      }
    }
    return true;
  }

  public static void main(String[] args) throws Throwable {  
    if (args.length != 2) {
      System.out.println("Args: [input file] [output file]");
      return;
    }
    Assembler assembler = new Assembler();
    if (assembler.readFile(args[0])) {
      assembler.compile(args[1]);
    } else {
      System.out.println("Compilation failed.");
    }
  }
  
}
