/*
 * TOFU Circuit Simulator
 * Copyright (C) 2007 Michael Birken
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
 * 
 * http://www.meatfighter.com/tofu
 */

package tofu.harryporter;

import java.io.*;

public class Emulator {

  private int[] memory = new int[32 * 1024];
  private int A, B, C, D, M, XY, J, PC;
  private boolean carry, zero, sign;
  private boolean advancePrinter;
  
  private void emulate() {
    boolean halt = false;
    while(!halt) {
      int I = memory[PC];
      PC = 0x7FFF & (PC + 1);
      
      if ((I&0xC0) == 0x00) { // MOV and CLR
        int source = 0;
        if ((I&7) != ((I>>3)&7)) {
          switch(I & 7) {
            case 0:
              source = A;
              break;
            case 1:
              source = B;
              break;
            case 2:
              source = C;
              break;
            case 3:
              source = D;
              break;
            case 4:
              source = getHigh(M);
              break;
            case 5:
              source = getLow(M);
              break;
            case 6:
              source = getHigh(XY);
              break;
            case 7:
              source = getLow(XY);
              break;
          }
        }
        switch((I>>3) & 7) {
          case 0:
            A = source;
            break;
          case 1:
            B = source;
            break;
          case 2:
            C = source;
            break;
          case 3:
            D = source;
            break;
          case 4:
            M = (source << 8) | getLow(M);
            break;
          case 5:
            M = (getHigh(M) << 8) | source;
            break;
          case 6:
            XY = (source << 8) | getLow(XY);
            break;
          case 7:
            XY = (getHigh(XY) << 8) | source;
            break; 
        }
      } else if ((I&0xF8) == 0x90) { // LD 
        int address = M & 0x7FFF;
        switch(I & 3) {
          case 0:
            A = memory[address];
            break;
          case 1:
            B = memory[address];
            break;
          case 2:
            C = memory[address];
            break;
          case 3:
            D = memory[address];
            break;
        }
      } else if ((I&0xF8) == 0x98) { // STO 
        if (M == 0xFFFE) {
          advancePrinter = true;
        } else if (advancePrinter && M == 0xFFFF) {
          advancePrinter = false;
          switch(I & 3) {
            case 0:
              System.out.print((char)A);
              break;
            case 1:
              System.out.print((char)B);
              break;
            case 2:
              System.out.print((char)C);
              break;
            case 3:
              System.out.print((char)D);
              break;
          }
        } else {
          int address = M & 0x7FFF;
          switch(I & 3) {
            case 0:
              memory[address] = A;
              break;
            case 1:
              memory[address] = B;
              break;
            case 2:
              memory[address] = C;
              break;
            case 3:
              memory[address] = D;
              break;
          }
        }
      } else if ((I & 0xF0) == 0x80) { // ALU
        int result = 0;
        switch(I & 7) {
          case 0:
            result = B + C;
            break;
          case 1:
            result = B + 1;
            break;
          case 2:
            result = B & C;
            break;
          case 3:
            result = B | C;
            break;
          case 4:
            result = B ^ C;
            break;
          case 5:
            result = B ^ 0xFF;
            break;
          case 6:
            result = (B << 1) | (B >> 7);
            break;
        }
        carry = (result & 0xFFFFFF00) > 0;
        sign = (result & 0x80) > 0;
        result = result & 0xFF;
        zero = result == 0;        
        switch((I >> 3) & 1) {
          case 0:
            A = result;
            break;
          case 1:
            D = result;
            break;
        }
      } else if (I == 0xB0) { // INC XY
        XY = 0xFFFF & (XY + 1);
      } else if ((I & 0xF1) == 0xA0) { // RETURN and HLT
        int source = 0;
        switch((I >> 1) & 3) {
          case 0:
            source = M;
            break;
          case 1:
            source = XY;
            break;
          case 2:
            source = J;
            break;
          case 3:
            halt = true;
            break;
        }
        switch((I >> 3) & 1) {
          case 0:
            XY = source;
            break;
          case 1:
            PC = source;
            break;
        }
      } else if ((I & 0xC0) == 0xC0) { // 16-bit instruction
        int immediate = memory[PC];
        PC = 0x7FFF & (PC + 1);
        immediate <<= 8;
        immediate |= memory[PC];
        PC = 0x7FFF & (PC + 1);
        
        if (isBit(I, 5)) {
          J = immediate;
        } else {
          M = immediate;
        }
        
        if ((isBit(I, 4) && sign)
            || (isBit(I, 3) && !carry)
            || (isBit(I, 2) && zero)
            || (isBit(I, 1) && !zero)) {
          if (isBit(I, 0)) {
            XY = PC;
          }
          PC = immediate;
        }
      } else if ((I & 0xC0) == 0x40) { // SET-8
        int value = I & 0x1F;
        if (isBit(value, 4)) {
          value = value | 0xE0;
        }
        if (isBit(I, 5)) {
          B = value;
        } else {
          A = value;
        }
      }
    }
  }
  
  private void print() {
    print(" A", A, 8);
    print(" B", B, 8);
    print(" C", C, 8);
    print(" D", D, 8);
    print(" J", J, 16);
    print(" M", M, 16);
    print("PC", PC, 16);
    print("XY", XY, 16);
    //System.out.printf("ZCS: %d%d%d%n", 
    //    zero ? 1 : 0, carry ? 1 : 0, sign ? 1 : 0);
  }
  
  private void print(String name, int r, int len) {
    String bin = Integer.toBinaryString(r);
    while(bin.length() < len) {
      bin = "0" + bin;
    }    
    String hex = Integer.toHexString(r).toUpperCase();
    while(hex.length() < (len >> 2)) {
      hex = "0" + hex;
    }
    System.out.printf("%s: %s %s %d %d%n", name, bin, hex, r, 
        len == 8 ? (byte)r : (short)r);    
  }
  
  private int getHigh(int word) {
    return 0xff & (word >> 8);
  }
  
  private int getLow(int word) {
    return 0xff & word;
  }
  
  private boolean isBit(int word, int bit) {
    return (word & (1 << bit)) > 0;
  }
  
  public void run(String fileName) {
    File file = new File(fileName);
    if (!file.exists()) {
      System.out.printf("Can't find file %s.%n", fileName);
      return;
    }
    if (file.length() > memory.length) {
      System.out.println("Input file > 32K.");
      return;
    }
    
    try {
      loadFile(file);
    } catch(IOException e) {
      System.out.printf("Error: %s%n", e.getMessage());
      return;
    }
    
    emulate();
    print();
  }
  
  private void loadFile(File file) throws IOException {
    BufferedInputStream bis = null;
    int b = 0;
    int index = 0;
    try {
      bis = new BufferedInputStream(new FileInputStream(file));
      while((b = bis.read()) >= 0) {
        memory[index] = b;
        index++;
      }
    } finally {
      if (bis != null) {
        bis.close();
      }
    }
  }
  
  public static void main(String[] args) {
    if (args.length != 1) {
      System.out.println("Args: [input file]");
      return;
    }
    Emulator emulator = new Emulator();
    emulator.run(args[0]);
  }
}
